from pathlib import Path

from yace.tools import Python3, Tool


class FooBarBaz(Tool):
    """The foobar tool which is nowhere to be found"""

    def __init__(self, cwd):
        super().__init__("foo_bar_baz", cwd)


def test_tool_does_not_exist():
    """When executables does not exists, then it exists() should failed"""

    foobar = FooBarBaz(Path.cwd())
    assert not foobar.exists()


def test_tool_invalid_usage():
    """When Passing invalid arguments run() should give a non-zero rcode"""

    python = Python3(Path.cwd())
    rcode, proc = python.run(["--foo"])

    assert rcode



"""
Tools used by targets to do what they do. The tool-classes are wrapper-classes
around executables. A brief overview:

* :class:`.Black`
* :class:`.ClangFormat`
* :class:`.Doxygen`
* :class:`.Gcc`
* :class:`.Isort`

The tools listed above can be classified in the following two categories.

Code Formaters
==============

Tools such as source-code formaters (clang-format, black, isort, rustfmt),
these will change the given code according to style-definitions. When these
tools change files, then they typically terminate using a non-zero exit-code.
This is not an error as it is commonly what we wanted the tool to do, that is a
target emits format and invokes a format-tool to make it conform.

Thus the tool is usually invoked twice, once for formating, and a second time
for checking that the code is well-formatted.

The above tools needs to be installed on the system. These classes inherit the
``run()`` from the :class:`.Tool`. Invoking a system :class:`.Tool` raises
:class:`.ToolError`.

Compilers
=========

Compilers such as :class:`.Gcc` usually terminate with a non-zero exit-code
when the input-program / source is invalid. Re-running the compiler does not
change this, the program needs to be changed. Something is wrong, and the user
needs to fix it.

Interpreters
============

...
"""
import logging as log
import typing
from pathlib import Path
from subprocess import STDOUT, run


class Tool(object):
    """
    Wrapper-class for invoking system tools
    """

    def __init__(self, executable, cwd):
        self.executable = executable
        self.cwd = cwd

    def run(self, args: typing.List[str]):
        """
        Invoke subprocess.run([self.executable] + args, ...); logging stdout
        and stderr to the given 'logfile' using arguments and self.cwd as
        cwd.

        On success, proc is returned.
        """

        logpath = (Path(self.cwd) / f"{self.executable}.log").resolve()
        cmd = [self.executable] + args

        with logpath.open("a") as logfile:
            logfile.write(f"# cmd({' '.join(cmd)})\n")
            logfile.write(f"# cwd({self.cwd})\n")
            logfile.flush()

            proc = run(
                cmd,
                stdout=logfile,
                stderr=STDOUT,
                check=False,
                capture_output=False,
                cwd=self.cwd,
            )

        rcode = proc.returncode
        if rcode:
            log.error(
                f"cmd({' '.join(cmd)}) exited with rcode({rcode}),"
                f" see logfile({logfile.name}) for details"
            )

        return rcode, proc

    def exists(self):
        """Returns true if the tool exists."""

        try:
            rcode, _ = self.run(["--version"])
            return rcode == 0
        except FileNotFoundError:
            log.error("executable(%s); FileNotFound", self.executable)

        return False


class Black(Tool):
    """
    Wrapper for the system-tool ``black``, usually utilized to format code
    emitted by Python-targets
    """

    def __init__(self, cwd):
        super().__init__("black", cwd)


class ClangFormat(Tool):
    """
    Wrapper for ``clang-format``
    """

    CLANGFORMAT_STYLE_C = "clang-format-c.clang-format"  # Style for .c files
    CLANGFORMAT_STYLE_H = "clang-format-h.clang-format"  # Style for .h files
    CLANGFORMAT_BIN = "clang-format"  # Name of the clang-format binary

    def __init__(self, cwd):
        super().__init__(ClangFormat.CLANGFORMAT_BIN, cwd)


class Doxygen(Tool):
    """
    Wrapper for ``doxygen``

    Run 'doxygen' using :attr:`.Doxygen.DOXYGEN_CONF`
    """

    DOXYGEN_CONF = "doxygen.conf"  # Filename of the Doxygen configuration file
    DOXYGEN_BIN = "doxygen"  # Name of the doxygen binary / executable

    def __init__(self, cwd):
        super().__init__(Doxygen.DOXYGEN_BIN, cwd)


class Gcc(Tool):
    """
    Wrapper for ``gcc``
    """

    def __init__(self, cwd):
        super().__init__("gcc", cwd)


class Isort(Tool):
    """
    Wrapper for ``isort``
    """

    def __init__(self, cwd):
        super().__init__("isort", cwd)


class Python3(Tool):
    """
    Wrapper for ``python3``
    """

    def __init__(self, cwd):
        super().__init__("python3", cwd)



from itertools import product
from pathlib import Path

import pytest

from yace.compiler import Compiler

VALID = [
    Path("models") / "example.yaml",
    Path("models") / "example-mini.yaml",
    Path("models") / "nvme.yaml",
    Path("models") / "xnvme.yaml",
]

INVALID = [
    Path("models") / "example-invalid.yaml",
]


@pytest.mark.parametrize("path,target", product(VALID, Compiler.TARGETS))
def test_compiler_with_valid_model(path, target):
    """Test **yace** via the cli using valid **yims**"""

    yace = Compiler([target.NAME], Path("/tmp") / "foo")
    yace.process(path)


@pytest.mark.parametrize("path,target", product(INVALID, Compiler.TARGETS))
def test_compiler_with_invalid_model(path, target):
    """Test **yace** via the cli using invalid **yims**"""

    yace = Compiler([target.NAME], Path("/tmp") / "foo")
    yace.process(path)



#!/usr/bin/env python3
"""
The **yace** data types consists of:

* No type: :class:`.Void`, :class:`.VoidPtr`

* Boolean: :class:`.Bool`

* Textual: :class:`.Char`

* Size types: :class:`.USize` / :class:`.ISize`

* Integers (Signed)

  * :class:`.I`, :class:`.I16`, :class:`.IHalf`, :class:`.I8`
  * :class:`.ILongLong`, :class:`.I64`, :class:`.ILong`, :class:`.I32`

* Integers (Unsigned)

  * :class:`.U`, :class:`.U16`, :class:`.UHalf`, :class:`.U8`
  * :class:`.ULongLong`, :class:`.U64`, :class:`.ULong`, :class:`.U32`

* Floating Point: :class:`.F32`, :class:`.F64`

Conventions
-----------

* A specialized Typespec should only update attribute-values, that is,
  not introduce new attributes

* A specialized Typespec should only set attributes which have values different
  than the default.

Their idl representation follows below.
"""
import inspect
import sys
import typing

from .base import Typespec


class Void(Typespec):
    """A void, that is, the type signaling no type"""

    cls: str = "void"

    void: typing.Optional[bool] = True


class VoidPtr(Typespec):
    """A void-pointer, that is, point to anything (including nothing)"""

    cls: str = "void_ptr"

    void: typing.Optional[bool] = True
    pointer: int = 1


class Char(Typespec):
    """
    A character; at least 8 bits wide.

    The C API emitter could produce::

        char

    for the :class:`.Char` entity.
    """

    cls: str = "char"

    character: bool = True
    width: int = 8


class ISize(Typespec):
    """
    Signed Size-type

    The C API emitter could produce::

        ssize_t

    for the :class:`.ISize` entity.
    """

    cls: str = "isize"

    size: bool = True
    width: typing.Optional[int] = 16


class I8(Typespec):
    """
    Signed integer exactly 8 bits wide.

    A C emitter could for example produce::

        int8_t

    for the :class:`.I8` entity.
    """

    cls: str = "i8"

    integer: bool = True
    width: int = 8
    width_fixed: bool = True


class IHalf(Typespec):
    """
    Signed integer at least 8 bits wide.

    This is also known as "half" or "short" int, as it is relative to "Int"
    shorter, typically, half the amount of bits.

    The C emitter can produce::

        short int

    for the :class:`.IHalf` entity.
    """

    cls: str = "ih"

    integer: bool = True
    width: typing.Optional[int] = 8


class I16(Typespec):
    """
    Signed integer exactly 16 bits wide.

    A C emitter could produce::

        int16_t

    for the :class:`.I16` entity.
    """

    cls: str = "i16"

    integer: bool = True
    width: typing.Optional[int] = 16
    width_fixed: bool = True


class I(Typespec):
    """
    Signed integer at least 16 bits wide.

    Historically, the int matched the bit-width of the processor when C came
    out in the early days on "ANSI" C / K&R. That is, 16bit processors.

    Thus, the short/long/long/long are "modifiers" of this base-type, although
    the width on a 64bit system today is often 32bit for this type.

    The C emitter can produce::

        int

    for the :class:`.I` entity.
    """

    cls: str = "i"

    integer: bool = True
    width: typing.Optional[int] = 16


class I32(Typespec):
    """
    Signed integer exactly 32 bits wide.

    A C emitter could produce::

        int32_t

    for the :class:`.I32` entity.
    """

    cls: str = "i32"

    integer: bool = True
    width: int = 32
    width_fixed: bool = True


class ILong(Typespec):
    """
    Signed integer at least 32 bits wide.

    The C emitter produces::

        long int

    for the :class:`.ILong` entity.
    """

    cls: str = "il"

    integer: bool = True
    width: typing.Optional[int] = 32


class I64(Typespec):
    """
    Signed integer exactly 64 bits wide.

    A C emitter could produce::

        int64_t

    for the :class:`.I64` entity.
    """

    cls: str = "i64"

    integer: bool = True
    width: typing.Optional[int] = 64
    width_fixed: bool = True


class ILongLong(Typespec):
    """
    Signed integer at least 64 bits wide.

    A C emitter could produce::

        long long int

    for the :class:`.ILongLong` entity.
    """

    cls: str = "ill"

    integer: bool = True
    width: typing.Optional[int] = 64


class USize(Typespec):
    """
    Unsigned Size-type

    A C emitter could produce::

        size_t

    for the :class:`.USize` entity.
    """

    cls: str = "usize"

    size: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 16


class U8(Typespec):
    """
    Unsigned integer exactly 8 bits wide.

    A C emitter could produce::

        uint8_t

    for the :class:`.U8` entity.
    """

    cls: str = "u8"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 8
    width_fixed: bool = True


class UHalf(Typespec):
    """
    Unsigned integer at least 8 bits wide.

    This is also known as "half" or "short" int, as it is relative to "Int"
    shorter, typically, half the amount of bits.

    A C emitter could produce::

        unsigned short int

    for the :class:`.UHalf` entity.
    """

    cls: str = "uh"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 8


class U16(Typespec):
    """
    Unsigned integer exactly 16 bits wide.

    A C emitter could produce::

        uint16_t

    for the :class:`.U16` entity.
    """

    cls: str = "u16"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 16
    width_fixed: bool = True


class U(Typespec):
    """
    Unsigned integer at least 16 bits wide.

    The C emitter can produce::

        unsigned int

    for the :class:`.U` entity.
    """

    cls: str = "u"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 16


class U32(Typespec):
    """
    Unsigned integer exactly 32 bits wide.

    A C emitter could produce::

        uint32_t

    for the :class:`.U32` entity.
    """

    cls: str = "u32"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 32
    width_fixed: bool = True


class ULong(Typespec):
    """
    Unsigned integer at least 32 bits wide.

    A C emitter could produce::

        unsigned long int

    for the :class:`.ULong` entity.
    """

    cls: str = "ul"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 32


class U64(Typespec):
    """
    Unsigned integer exactly 64 bits wide.

    A C emitter could produce::

        uint64_t

    for the :class:`.U64` entity.
    """

    cls: str = "u64"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 64
    width_fixed: bool = True


class ULongLong(Typespec):
    """
    Unsigned integer at least 64 bits wide.

    A C emitter could produce::

        unsigned long long int

    for the :class:`.ULongLong` entity.
    """

    cls: str = "ull"

    integer: bool = True
    unsigned: bool = True
    width: typing.Optional[int] = 64


class F32(Typespec):
    """
    Floating point numerical value, possibly 32 bits wide

    The C language spec. does not define the width to be exact. However, we
    treat it as though it is, and error if it is not.

    A C emitter could produce::

        float

    for the :class:`.F32` entity.
    """

    cls: str = "f32"

    real: bool = True
    width: int = 32
    width_fixed: bool = True


class F64(Typespec):
    """
    Floating point numerical value, possibly 64 bits wide

    The C language spec. does not define the width to be exact. However, we
    treat it as though it is, and error if it is not.

    A C emitter could produce::

        double

    for the :class:`.F64` entity.
    """

    cls: str = "f64"

    real: bool = True
    width: int = 64


class Bool(Typespec):
    """
    A boolean, at least 8 bits wide

    Equivalent to the C99 "_Bool" and available
    as "bool" as defined by the standardized "stdbool.h" header.

    A C emitter could produce::

        # In C99
        _Bool

    Or, using ``<stdbool.h>``::

        bool

    for the :class:`.Bool` entity.
    """

    cls: str = "bool"

    boolean: bool = True
    width: int = 8


class String(Typespec):
    """
    A string pointer

    The C emitter produces::

        const char *

    for the :class:`.String` entity.
    """

    cls: str = "string"

    character: bool = True
    pointer: int = 1
    const: bool = True


def classes():
    """Return all datatype classes"""

    return [
        cls
        for _, cls in inspect.getmembers(sys.modules[__name__])
        if inspect.isclass(cls) and issubclass(cls, Typespec) and not cls == Typespec
    ]


def classes_shorthand_data():
    """Returns a map of shorthands to non-default datatype data"""

    default = Typespec({"cls": "typespec"}).as_dict()  # Default typespec

    shorthands = {}  # Map to return
    for cls in classes():
        data = cls().as_dict()  # Setup data
        del data["lbl"]  # Remove "lbl"

        for key, val in default.items():  # Remove default-values
            if key in data and data.get(key) == val and key not in ["cls"]:
                del data[key]

        shorthands[cls.cls] = data  # Add to shorthands

    return shorthands



#!/usr/bin/env python3
"""
The :class:`.Model` provides the root-node for the **Yace** Interface
Definition, it provides helper-functions for loading **Yace** Interface
Definition Language files from their ``.yaml`` format on the file-system, into
the representation offered by the **Yace** Interface Definition
Language-entities.

* :class:`.Meta`
* :class:`.Model`
* :class:`.ModelWalker`

Their **Yace** Interface Definition Language representation follows below.
"""
import inspect
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple

import yaml

from yace.errors import InvalidModelData
from yace.idl import base, constants, datatypes, derivedtypes, functiontypes


def data_from_yaml(path: Path) -> Tuple[Dict, List[Dict]]:
    """
    Returns data from YAML (yim) file as (meta, entities)
    """

    meta = {}
    entities = []

    with path.open() as yamlfile:
        for lbl, data in yaml.safe_load(yamlfile).items():
            if lbl == "meta":
                meta = data
                continue

            for entity in data:
                if "lbl" not in entity:
                    entity["lbl"] = []
                entity["lbl"].append(lbl)
                entities.append(entity)

    return meta, entities


@dataclass
class Meta:
    """
    Meta data used by the to describe attributes of the generated C API and
    bindings/wrappers, such as license, version, and documentation
    """

    lic: str  # License string, this is utilized for SPDX
    version: str = "0.0.1"
    author: str = "Foo Bar <foo@example.com>"
    project: str = "Example"
    prefix: str = "ex"
    brief: str = "yace example"
    full: str = (
        "This description is provided with the default idl-meta, to "
        "change it, then make sure you have a meta-lbl in your yace-file"
    )


class Model(object):
    """
    The :class:`.Model` serves as the root of the "tree" describing
    the interface to generate C API and bindings for.
    """

    MAPPING = {
        obj.cls: obj
        for _, obj in (
            inspect.getmembers(base)
            + inspect.getmembers(constants)
            + inspect.getmembers(datatypes)
            + inspect.getmembers(derivedtypes)
            + inspect.getmembers(functiontypes)
        )
        if inspect.isclass(obj) and hasattr(obj, "cls")
    }

    def __init__(self):
        self.meta = Meta(lic="UNLICENSED")
        self.entities = []

    @classmethod
    def entity_from_data(cls, cur: dict, parent=None, depth=0):
        """
        Walk 'cur' recursively with depth-count and "short-hand" expansion

        A tree-walk is needed as just invoking the dataclass-constructor::

            entity = Model.MAPPING["struct"](data)

        Would not instantiate members, parameters, typ, etc. additionally,
        the short-hands are expanded:

        - Integer literals

          - {typ: 10} instead of {typ: {cls: dec, val: 10}}

        - Data Types

          - {typ: u16} instead of {typ: {cls: u16}}

        Foo...
        """

        if type(cur) not in [int, str, dict]:
            raise InvalidModelData(f"Invalid type: {cur}")
        elif type(cur) in [dict] and "cls" not in cur:
            raise InvalidModelData(f"Missing 'cls' in {cur}")

        if isinstance(cur, int):  # short-hand for integer-literal
            return constants.Dec({"lit": cur})
        elif isinstance(cur, str):  # short-hand for typ
            typ = Model.MAPPING.get(cur)
            if not typ:
                raise InvalidModelData(f"typ !short-hand: '{cur}'")

            return typ({})

        constructor = Model.MAPPING.get(cur["cls"])
        if not constructor:
            raise InvalidModelData(f"No constructor for '{cur}'")

        attributes = {}
        for attr, attr_data in cur.items():
            if attr in ["typ", "ret", "val"]:
                attributes[attr] = Model.entity_from_data(attr_data, cur, depth)
            elif attr in ["members", "parameters"]:
                attributes[attr] = [
                    Model.entity_from_data(child, cur, depth + 1)
                    for child in cur.get(attr)
                ]
            else:
                attributes[attr] = attr_data

        return constructor(attributes)

    @classmethod
    def from_data(cls, meta: dict, ents: list):
        """Construct a :class:`Model` using the given 'meta' and 'entities'"""

        interface = cls()
        interface.meta = Meta(**meta)

        total = len(ents)
        for count, entity_data in enumerate(ents, 1):
            logging.debug(f"Processing {count} / {total} in {entity_data['lbl']}")
            entity = cls.entity_from_data(entity_data)
            interface.entities.append(entity)

        return interface

    @classmethod
    def from_path(cls, path: Path):
        """
        Returns a dict composed of the merged content of all yaml-files in the
        given 'path'
        """

        return cls.from_data(*data_from_yaml(path))


class ModelWalker(object):
    """
    Base-class for walking the :class:`.Model`

    This is useful for inspecting properties of the model.

    **HOWTO**: sub-class this and implement the visit() method
    """

    def __init__(self, model):
        self.model = model

    def _traverse(
        self,
        cur,
        ancestors: List,
        depth: int = 0,
    ) -> str:
        """
        Recursive walk of the 'cur' entity, returning a list of visit()
        results.
        """

        res = [self.visit(cur, ancestors, depth)]
        for attr in ["members", "parameters", "typ", "ret", "val"]:
            other = getattr(cur, attr, None)
            if other is None:
                continue

            if attr in ["members", "parameters"]:
                for child in other:
                    res += self._traverse(child, ancestors + [cur], depth + 1)
            else:
                res += self._traverse(other, ancestors + [cur], depth + 1)

        return res

    def walk(self):
        """Walks the :class:`.Model` invoking visit()"""

        status = []
        for entity in self.model.entities:
            status += self._traverse(entity, [], 0)

        return status

    def visit(self, current, ancestors, depth):
        """This is the thing which class should implement"""

        return (True, None)



"""
YAML ==> Interface(List[Entities]) ==> Emitter() ==> CodeTarget
"""
from pathlib import Path

from jinja2 import Environment, PackageLoader


def camelcase(symbol, pascalcase=True):
    """Format the given 'symbol' as (C|c)amelCase"""

    camelized = "".join(
        [part[0].upper() + part[1:].lower() for part in symbol.split("_")]
    )
    if pascalcase:
        return camelized

    return camelized[0].lower() + camelized[1:]


class Emitter(object):
    """
    The default approach to transforming :class:`.Entity` in :class:`.Model`:
    to source-code-text is inheriting **this** class and utilizing the Jinja
    templates made available in self.templates. It might seem naive to utilize
    templating-engine like Jinja for a code-emitter, however, so far it seems
    to be the an incredibly simple approach.

    This class populates ``self.templates`` with templates loaded for a given
    emitter-implementation. They work as Jinja usually does, a feature-addition
    allowing you to render a Jinja template **within** a Jinja template, via a
    custom-filter-function. For example, like so::

        {{ entity | emit_entity }}

    And available::

        {{ entity | emit_typespec }}

    This works by seperating Jinja templates by filename-convention,
    template-names starting with "entity_{name}.template" are usable by this
    in-template renderer. Thus, for the example above, the template file is
    named ``entity_typedecl``.

    It assumes that 'entity' is an instance of :class:`.Entity` and uses a
    Jinja template to render it. This behavior is added as a means to keep
    **all** text-formating code inside the Jinja-templates, one does not need
    to goto the Python to see a different technique. Thus, **all**::

        Entity ==> SourceCodeText

    Is by default handled in one place: the Jinja templates.
    """

    def __init__(self, output: Path, name: str = "base"):
        self.name = name
        self.output = output.resolve()

    def render(self, template, args):
        """Renders the given template, passing args..."""

        typespec_jenv = Environment(
            loader=PackageLoader(f"yace.targets.{self.name}", "."),
            extensions=["jinja2.ext.do"],
        )
        typespec_jenv.globals.update(zip=zip, len=len)
        typespec_templates = {
            Path(f).stem.split(".")[0]: typespec_jenv.get_template(f)
            for f in typespec_jenv.list_templates()
            if f.endswith(".template") and f.startswith("typespec")
        }

        def emit_typespec(entity, anon: bool = False):
            args["entity"] = entity
            template = "typespec_anon" if anon else "typespec"

            return typespec_templates[template].render(**args)

        entity_jenv = Environment(
            loader=PackageLoader(f"yace.targets.{self.name}", "."),
            extensions=["jinja2.ext.do"],
        )
        entity_jenv.globals.update(zip=zip, len=len)
        entity_jenv.filters["camelcase"] = camelcase
        entity_jenv.filters["emit_typespec"] = emit_typespec
        entity_templates = {
            Path(f).stem: entity_jenv.get_template(f)
            for f in entity_jenv.list_templates()
            if f.endswith(".template") and f.startswith("entity_")
        }

        def emit_entity(entity, depth: int = 0):
            args["entity"] = entity
            args["depth"] = depth

            return entity_templates[f"entity_{entity.cls}.h"].render(**args)

        file_jenv = Environment(
            loader=PackageLoader(f"yace.targets.{self.name}", "."),
            extensions=["jinja2.ext.do"],
        )
        file_jenv.globals.update(zip=zip, len=len)
        file_jenv.filters["camelcase"] = camelcase
        file_jenv.filters["emit_typespec"] = emit_typespec
        file_jenv.filters["emit_entity"] = emit_entity
        file_templates = {
            Path(f).stem: file_jenv.get_template(f)
            for f in file_jenv.list_templates()
            if f.endswith(".template") and not f.startswith("entity_")
        }

        return file_templates[template].render(**args)



#!/usr/bin/env python3
"""
This module provides helpers to format a **Yace**-file, that is, a
``.yaml``-file with Yace Interface Definition entities. A formater provides a
tool-based approach to maintain a canonical format for **Yace**-files.

Since there is a lot of flexibility in **.yaml**-files, a canonical format makes it
easier to read what others have written. Here is what the canonical format
entails:

* The document starts with ``---``

* The ``meta``-key is the first key in the document

  * Raionale: when reading a **Yace**-file, the meta-data should provide the
    introduction to the **Yace**-file, and thus be the first thing int the document

* Top-level keys has an empty line above it

  * Rationale: the top-level keys in **Yace**-file have no significance except
    for the 'meta' key, the intent of all other keys is to provide a "header" /
    "topic" for the list of entities 

* Entity-keys are sorted by an explicit :class:`yace.idl.formater.ORDER`:

  * Rationale: this is done for consistency, always provide a well-known
    ordering, makes it easier to read **Yace**-file files written by others
  * See the :class:`.ORDER` for the explicit ordering used

* Print "flattened", when there are no nested nested structures

  * Rationale: this is typically function-arguments or field-members, as such,
    one would usually read these in a flattened form

Example::

  foo: {cls: dec, sym: min_x, val: 128, doc: Foo Bar Baz}

* Prints block, entities in a block they contain nested values

Example::

  foo:
    cls: dec
    sym: min_x
    val: 128
    doc: Foo Bar Baz

Ideally, then the ``yaml.dump()``, could be parametetrized, or some other
tool/api could do the above. However, I have not found anything which could.

Thus this module. In case somebody reads this, and knows of a simpler approach
achieving the above, please to send a PR.
"""
from pathlib import Path
import typing
import string
import yaml
import os
import logging as log

ORDER = [
    "meta",
    "cls",
    "sym",
    "typ",
    "val",
    "width",
    "width_fixed",
    "fmt",
    "doc",
    "project",
    "version",
    "brief",
    "full",
    "prefix",
    "author",
    "lic",
]


def reorder(data, order):
    """Manipulate the given 'data' adjusting key-order according to 'order'"""

    new = dict()
    for key in (key for key in order if key in data):
        new[key] = data[key]
        del data[key]
    for key in list(data.keys()):
        new[key] = data[key]
        del data[key]

    new.update(data)
    data.update(new)


def ydata_from_file(path: Path) -> dict:
    """Returns a yaml-file in without any transformation"""

    with path.open() as yfile:
        return yaml.safe_load(yfile)


def hascontainer(items):
    """Checks whether a dict or list is in the given list of items"""

    for item in items:
        if isinstance(item, (dict, list)):
            return True

    return False


def yfmt(data, depth):
    """Format the given data"""

    content = ""

    if data is None:
        return "null"

    elif isinstance(data, bool):
        return f"{data}".lower()

    elif isinstance(data, (bool, int, float)):
        return f"{data}"

    elif isinstance(data, str):
        """Render a string..."""

        # TODO: add support for
        # * multi-line strings
        # * Auto-detect whether ' or " should be used

        if not data:
            return "''"

        quote = ""
        if list(set(data) - set(string.ascii_letters + string.digits + "._- ")):
            quote = "'"

        return f"{quote}{data}{quote}"

    elif isinstance(data, dict):
        if depth:
            reorder(data, ORDER)

        # Flat / single-line
        if not hascontainer(list(data.values())):
            inner = ", ".join([f"{key}: " + yfmt(val, 0) for key, val in data.items()])
            return "{" + f"{inner}" + "}"

        # Block / multi-line
        lines = []
        for index, (key, val) in enumerate(data.items()):
            prefix = " " * int(depth // 2) * 2 if index else ""
            lines.append(f"{prefix}{key}: " + yfmt(val, depth + 1))

        linesep = "\n" if depth else "\n\n"
        content += linesep.join(lines)

    elif isinstance(data, list):
        lines = []
        for val in data:
            prefix = " " * int(depth // 2) * 2 + "- "
            lines.append(prefix + yfmt(val, depth + 1))

        content += "\n" + "\n".join(lines)

    else:
        print("[[[%s]]]", data)
        content += str(data)

    return content


def ydata_to_file(ydata: dict, path: Path):
    """Dump Yace-data to file"""

    with path.open("w") as file:
        file.write("---\n")

        meta = ydata.pop("meta", None)  # To ensure "meta" is the first key
        if meta:
            reorder(meta, ORDER)
            file.write(
                yaml.safe_dump(
                    {"meta": meta}, default_flow_style=False, sort_keys=False
                )
            )
            file.write("\n")

        file.write("\n".join([line.rstrip() for line in yfmt(ydata, 0).splitlines()]))
        file.write("\n")


def do_format(paths: typing.List[Path]):
    """Format a list of paths in-place"""

    for path in paths:
        path_backup = path.resolve().with_name(f"{path.name}.bck")
        ydata = ydata_from_file(path)
        ydata_to_file(ydata, path_backup)

        # Open and compare
        old = ydata_from_file(path)
        new = ydata_from_file(path_backup)
        if old != new:
            log.error(f"Content altered, aborting. See({path}, {path_backup})")
            return 1

        os.replace(path_backup, path)

    return 0



#!/usr/bin/env python3
"""
YACE raises exceptions in case of error.

The errors are usually il-formated YAML-files, incorrect
:class:`yace.idl.interface.Model`, type errors in the constructed
:class:`yace.idl.interface.Model`, layout violations in the
:class:`yace.idl.interface.Model`.
"""


class TransformationError(Exception):
    """Something went wrong during a yid-transformation"""

    pass


class ToolError(Exception):
    """Exception raised when a tool exists with non-zero returncode"""

    pass


class InvalidModelData(Exception):
    """The data given to entity instantion is not valid"""

    pass



import logging as log

from yace.model import ModelWalker


class Linter(ModelWalker):
    """
    Language Integrity checker for **Yace** Interface Defintion Files
    """

    def visit(self, current, ancestors, depth: int):
        """
        Visits the entity / typedecl and returns the status is_valid() on it
        """

        return current.is_valid()

    def check(self):
        """
        Returns the number of integrity issues found checking checking the
        given idl. Any messages are logged.
        """

        nerrors = 0
        for is_valid, message in self.walk():
            if is_valid:
                continue

            nerrors += 1
            log.error("message(%s)", message)

        return nerrors



"""
The goal of this module is to provide the buildings blocks for a yace-file
generator. That is, something to aid the creation of **Yace**-files.

In this specific instance, to parse C Headers, and emit equivalent
**Yace**-files.

* :class:`.CParser`, parse a C header and emit a **Yace**-file
"""
import logging as log
import re
from pathlib import Path

import typing
from clang.cindex import Config, CursorKind, Index
import yace.idl.datatypes
from yace.idl.formater import ydata_to_file


REGEX_INTEGER_FIXEDWIDTH = "(?P<unsigned>u?)int(?P<width>8|16|32|64|128)_t"

SHORTHAND_DATA = yace.idl.datatypes.classes_shorthand_data()


def typespec_data_from_custor(cursor) -> dict:
    """
    Returns a typespec for the given cursor

    NOTE:

    * Assumes CursorKind in [PARAM_DECL, FIELD_DECL]
    * Array-Handling; replaces symbolic constants with numerical value

    It currently logs, but does not raise when:

    * Parsing lead to type confusion
    * Parsing hit unknown token

    TODO:

    * Handle bits and bitfields
    * Raise when const appears as anything but the first modifier
    * Raise after loggin error (type confusion and unknown token)
    """

    tokens_all = [tok.spelling for tok in cursor.get_tokens()]

    is_nested = not list(set(["{", "}"]) - set(tokens_all))
    if is_nested:
        return {"cls": "typespec", "doc": "Nested struct"}

    is_bits = not list(set([":"]) - set(tokens_all))
    if is_bits:
        return {"cls": "bits", "doc": "Bitfield"}

    data = {"cls": "typespec"}

    # Skipping the last token;
    #
    # For PARAM and FIELD, then the last token is the identifier and thus not
    # part of the type itself
    last_token_index = -1

    # Skipping the last three tokens as these should be the tokens: "[", "FOO", "]"
    array_len = cursor.type.get_array_size()
    if array_len > 0:
        data["array"] = cursor.type.get_array_size()

        if not (
            len(tokens_all) > 4 and tokens_all[-1] == "]" and tokens_all[-3] == "["
        ):
            log.error("fooo")

        last_token_index = -4

    # Process the non-skipped tokens
    tokens = tokens_all[:last_token_index]
    for index, token in enumerate(tokens, 1):
        match = re.match(REGEX_INTEGER_FIXEDWIDTH, token)
        if match:
            data["integer"] = True
            data["unsigned"] = bool(match.group("unsigned"))
            data["width"] = int(match.group("width"))
            data["width_fixed"] = True
        elif token in ["size_t"]:
            data["size"] = True
            data["unsigned"] = True
            data["width"] = 16
        elif token in ["ssize_t"]:
            data["size"] = True
            data["width"] = 16
        elif token in ["short"]:
            data["integer"] = True
            data["width"] = 8
        # Integer without width modifier
        elif token in ["int"]:
            data["integer"] = True
            if "long" not in tokens and "short" not in tokens:
                data["width"] = 16
        elif token in ["long"]:
            data["integer"] = True
            data["width"] = 32 * sum([tok == token for tok in tokens])
        elif token in ["signed"]:
            data["integer"] = True
        elif token in ["unsigned"]:
            data["integer"] = True
            data["unsigned"] = True
        elif token in ["struct", "union"]:
            data[token] = True
            data["sym"] = tokens[-1]
            break  # last token should be name, anything in between we don't care about
        elif token in ["_Bool", "bool"]:
            data["boolean"] = True
        elif token in ["char"]:
            data["character"] = True
            data["width"] = 8
        elif token in ["*"]:
            if "pointer" not in data:
                data["pointer"] = 0
            data["pointer"] += 1
        elif token in ["const", "static"]:
            data[token] = True
        elif token in ["float"]:
            data["real"] = True
            data["width"] = 32
            data["width_fixed"] = True
        elif token in ["double"]:
            data["real"] = True
            data["width"] = 64
            data["width_fixed"] = True
        elif token in ["void"]:
            data[token] = True
        else:
            log.error(
                "Unknown token: {'%s', %d/%d}, tokens: %s",
                token,
                index,
                len(tokens),
                tokens,
            )

    typecount = sum(
        [
            val is True
            for key, val in data.items()
            if key
            in ["void", "boolean", "character", "integer", "real", "union", "struct"]
        ]
    )
    if typecount != 1:
        log.error("Type confusion; typecount: %d", typecount)

    return data


def typespec_data_to_typ(data):
    """
    Some typespec instances have short-hands, these are used for 'typ'
    keys, this function produces the short or data-identify
    """

    ignore = ["cls", "lbl", "array"]
    for shorthand, shorthand_data in SHORTHAND_DATA.items():
        if {key: val for key, val in shorthand_data.items() if key not in ignore} != {
            key: val for key, val in data.items() if key not in ignore
        }:
            continue

        if "array" not in data:
            return shorthand

        # Compact array representation
        #
        # Remove the shorthand-attributes, as the assignment of the shorthand-class
        # carries the attributes, thereby providing the most compact representation
        [data.pop(key) for key in shorthand_data.keys() if key in data]
        data["cls"] = shorthand

        return data

    return data


class CParser(object):
    """
    Primitive wrapper around libclang Python bindings
    """

    def __init__(self):
        """Figure out a way to setup the index..."""

        Config().set_library_path("/Library/Developer/CommandLineTools/usr/lib/")
        index = Index.create()
        self.index = index

    def parse_file(self, path: Path):
        """Parse the given file into a :class:`clang.cindex.TranslationUnit`."""

        return self.index.parse(path)

    def parse_enum(self, cursor, data):
        """Parse into cls-objects enum and enum_value"""

        data["cls"] = "enum"
        data["members"] = []

        for child in cursor.get_children():
            if child.kind not in [CursorKind.ENUM_CONSTANT_DECL]:
                log.warning(f"Skipping: {child} not ENUM_CONSTANT_DECL")
                continue

            data["members"].append(
                {
                    "cls": "enum_value",
                    "sym": child.spelling,
                    "doc": child.brief_comment,
                    "val": child.enum_value,
                }
            )

    def parse_struct(self, cursor, data):
        data["cls"] = "struct"
        data["members"] = []

        for child in cursor.get_children():
            if child.kind not in [CursorKind.FIELD_DECL]:
                log.warning(f"Skipping: {child} not ENUM_CONSTANT_DECL")
                continue

            child_data = {
                "cls": "field",
                "sym": child.spelling,
                "doc": child.brief_comment,
                "typ": typespec_data_to_typ(typespec_data_from_custor(child)),
            }
            data["members"].append(child_data)

    def parse_union(self, cursor, data):
        """Parse union declaration"""

        data["cls"] = "union"

    def parse_fun(self, cursor, data):
        """Parse function declaration"""

        data["cls"] = "fun"
        data["ret"] = {
            "cls": "ret",
            "doc": "TODO: parse / extract return-doc",
        }
        data["parameters"] = []

        for child in cursor.get_children():
            if child.kind not in [CursorKind.PARM_DECL]:
                log.warning(f"Skipping: {child} not PARM_DECL")
                continue

            child_data = {
                "cls": "param",
                "typ": typespec_data_to_typ(typespec_data_from_custor(child)),
                "sym": child.spelling,
                "doc": child.brief_comment,
            }
            data["parameters"].append(child_data)

    def tu_to_data(self, tu):
        """Transform the given translation-unit (tu) to data"""

        entities = []
        for cursor in tu.cursor.get_children():
            entity = {
                "sym": cursor.spelling,
                "doc": cursor.brief_comment,
            }
            if cursor.kind in [CursorKind.ENUM_DECL]:
                self.parse_enum(cursor, entity)
            elif cursor.kind in [CursorKind.STRUCT_DECL]:
                self.parse_struct(cursor, entity)
            elif cursor.kind in [CursorKind.UNION_DECL]:
                self.parse_union(cursor, entity)
            elif cursor.kind in [CursorKind.FUNCTION_DECL]:
                self.parse_fun(cursor, entity)
            else:
                continue

            entities.append(entity)

        return entities


def c_to_yace(paths: typing.List[Path], output: Path):
    """Optimistically / best-offort transformation of a C Header to YACE File"""

    # TODO: this information / structure should be read from a single point,
    # somewhere in the yace.idl module
    ydata = {
        "meta": {
            "lic": "Unknown License",
            "version": "0.0.1",
            "author": "Foo Bar <foo@example.com>",
            "project": "foo",
            "prefix": "foo",
            "brief": "Brief Description",
            "full": "Full Description",
        },
    }
    ydata["entities"] = []

    parser = CParser()
    for path in [p.resolve() for p in paths]:
        tu = parser.parse_file(path)
        ydata["entities"] += parser.tu_to_data(tu)

        ydata_to_file(ydata, output / f"{path.stem}_parsed.yaml")

    return 0



import pytest

from yace.emitters import camelcase

SNAKECASE = ["foo", "foo_bar", "foo_bar_baz"]
PASCALCASE = ["Foo", "FooBar", "FooBarBaz"]
CAMELCASE = ["foo", "fooBar", "fooBarBaz"]


@pytest.mark.parametrize("symbol,expected", zip(SNAKECASE, PASCALCASE))
def test_camelize_pascalcase(symbol, expected):
    assert camelcase(symbol, True) == expected


@pytest.mark.parametrize("symbol,expected", zip(SNAKECASE, CAMELCASE))
def test_camelize_camelcase(symbol, expected):
    assert camelcase(symbol, False) == expected



#!/usr/bin/env python3
"""
The **yace** derived types consists of:

* :class:`yace.idl.derivedtypes.Struct` with :class:`.Field` members

* :class:`yace.idl.derivedtypes.Bitfield` with :class:`.Bits` members

* :class:`yace.idl.derivedtypes.Union`

Their idl representation follows below.
"""
import typing

from .base import Documented, Entity, Named, Typed


class Bits(Entity, Named, Documented):
    """Representation of a :class:`.Bitfield` member."""

    cls: str = "bits"
    width: int


class Bitfield(Entity, Named, Documented):
    """
    Representation of a bitfield, that is a partitioning of a fixed-width
    datatype (such as uint32_t) into a subset of named bit-ranges. Commonly
    encapsualted in C struct. For example::

        struct {
            uint8_t foo : 2;
            uint8_t bar : 3;
            uint8_t baz : 3;
        }

    By doing so, then a struct-accessor is provided at less-than
    byte-addressability, which is very convenient. By providing a semanticly
    rich representation in **yace** then the actual implementation in a given
    language can be something better fitted.

    For example, the above code in C is not portable, whereas other languages
    might support portable representations. By providing the abstract
    bit-field, then the code-emitter can generate what is best-suited for a
    given language, rather than translating C-representation-idioms as the
    above into something sub-optimal.
    """

    cls: str = "bitfield"
    width: int
    members: typing.List[Bits]

    def is_valid_members(self):
        """Checks whether the bitfield members match the width"""

        invalid_types = [type(m) for m in self.members if not isinstance(m, Bits)]
        if len(invalid_types):
            return (
                False,
                f"Invalid types in {self.sym}.members; "
                "expecting: Bits, got: [{invalid_types}]",
            )

        acc = sum([m.width for m in self.members])
        if acc != self.width:
            return (False, f"Aacumulated width({acc}) != {self.width}")

        return (True, None)


class Field(Entity, Named, Documented, Typed):
    """
    A representation of :class:`yace.idl.Struct` and
    :class:`yace.idl.Union`, utilized by compiler to emit code in the C API
    similar to::

        uint32_t bar;

    That is, within a ``struct/union`` block e.g.::

        struct foo {
            uint32_t bar;
        };

    The 'typ' and 'typ.width' members inherited from :class:`.Typed` are
    utilized to produce a fitting type-declaration for the target language, as
    in the examples above for C.

    The 'fmt' attribute provides a format-char/string for a code-emitter to
    utlize when emitting a pretty-printer.
    """

    cls: str = "field"
    fmt: str


class Struct(Entity, Named, Documented):
    """
    A representation of a struct definition
    """

    cls: str = "struct"
    members: typing.List[Field]


class Union(Struct):
    """Representation of enumerations / collections of constants"""

    cls: str = "union"



"""
The entities in the **yace** Interface Definition are enherit the
:class:`.Entity` and mix-in common attributes.

* Base-classes: :class:`.Entity` / :class:`.Typespec`

* Mixin-attributes: :class:`.Named` / :class:`.Documented` / :class:`.Typed`

The above are defined below.
"""
import logging as log
import re
import typing
from collections import ChainMap


def all_annotations(cls) -> ChainMap:
    """
    Returns a dictionary-like ChainMap that includes annotations for all
    attributes defined in the given 'cls' or inherited from superclasses.

    Courtesy of SO:
    https://stackoverflow.com/questions/63903901/how-can-i-access-to-annotations-of-parent-class
    """
    return ChainMap(
        *(c.__annotations__ for c in cls.__mro__ if "__annotations__" in c.__dict__)
    )


class Entity(object):
    """
    This is the baseclass for all entities in the Yace Interface Model
    (:class:`yace.model.Model`)

    It is defined as the Python dataclasses got in the way, this instead as a
    fixed set of functionality:

    * Construct by passing a dict of arguments
    * Do a runtime type using is_valid()
    * Allow inheritance / subclassing with optional and required arguments

    Constructor raises ValueError for missing required arguments.
    """

    cls: str  # The Yace-IDL constructor keyword
    lbl: typing.List[str] = ["nope"]  # Labels

    def __init__(self, args: dict = {}):
        self.annotations = all_annotations(self.__class__)

        self.all = list(self.annotations)
        self.required = []
        self.optional = []
        for attr in self.annotations.keys():
            if hasattr(self, attr):
                self.optional.append(attr)
            else:
                self.required.append(attr)

        missing = list(set(self.required) - set(list(args)))
        unexpected = list(set(args) - set(self.all))

        if missing or unexpected:
            log.info(f"ant: {self.annotations}")
            log.info(f"all: {self.all}")
            log.info(f"req: {self.required}")
            log.info(f"opt: {self.optional}")
            log.info(f"missing: {missing}")
            log.info(f"unexpected: {unexpected}")
            raise ValueError(
                f"{self.__class__.__name__}() constructor arguments; "
                f"is missing: {missing}, "
                f"got unexpected: {unexpected}"
            )

        for key, val in args.items():
            setattr(self, key, val)

    def is_valid(self):
        """
        Here is what this validates

        * Field-specific checks from subclass or mixin such as ``is_valid_sym()``

        * Generic check of attribute-instances matching theirannotations
        """

        for attr, atype in self.annotations.items():
            if not getattr(attr, "get_origin", False):  # skip typing.*
                continue
            if isinstance(attr, atype):  # is_valid
                continue

            return False, f"Bad type({atype}) of attribute({attr})"

        for attr in self.all:
            attr_checker = getattr(self, f"is_valid_{attr}", False)
            if not attr_checker:
                continue
            valid, message = attr_checker()
            if valid:
                continue

            return valid, message

        return True, "OK"

    def as_dict(self):
        """Returns the Entity represented as a dict"""

        return {attr: getattr(self, attr) for attr in self.all}


class Typespec(Entity):
    """
    All entities is-a :class:`.Entity`
    """

    # Lack of type
    void: bool = False

    # Boolean and textual datatypes
    boolean: bool = False  # Boolean type, since C99: _Bool / true / false
    character: bool = False  # Character type, in C: char / 'k'

    # Numeral types
    size: bool = False  # Size-types; size_t, ssize_t
    integer: bool = False  # Integer types; int
    real: bool = False  # Floating point number e.g. float, double

    # Numerical types (Qualifiers and Modifiers)
    unsigned: bool = False  # Type-modifier for the integer datatype
    width: typing.Optional[int] = None  # integer type width
    width_fixed: bool = False  # True: has an exact bit-width

    # Derived types
    union: bool = False  # union <id> { ... }
    struct: bool = False  # struct <id> { ... }
    sym: typing.Optional[str] = None  # symbol / tag of structs and unions

    # General Qualifiers and Modifiers
    const: bool = False  # Access-qualifier for all types:  'const'
    static: bool = False  # Storage-qualifier for all types: 'static'
    pointer: int = 0  # Pointer-type for all types: '*'
    array: int = 0  # Array-type for all types: '[]'


class Documented(object):
    """
    Attribute-mixin; adding a required 'doc' describing the :class:`.Entity`
    """

    doc: str


class Named(object):
    """
    Attribute-mixin; adding a required 'sym' to be used as identifier
    """

    sym: str

    def is_valid_sym(self):
        """Check whether the symbol is a valid C identifier"""

        if self.sym is None:
            return False, f"Invalid attr: sym == None; for {self.as_dict()}"

        match = re.match(r"[_a-z][_a-z0-9]{0,30}", str(self.sym))
        if match:
            return True, "OK"

        return False, f"Invalid attr: '{self.sym}'; for {self.as_dict()}"


class Typed(object):
    """
    Attribute-mixin; adding a reqried "has-a" relation to :class:`.Typespec`
    """

    typ: Typespec



#!/usr/bin/env python3
"""
The is the command-line interface to **yace**.

TODO
====

* Discover 'adhoc' targets, e.g. targets available in the
  current-workding-directory
* Discover targets via packages under the namespace 'yace.targets.{NAME}`

"""
import argparse
import logging as log
import sys
from pathlib import Path

from yace import __version__ as version
from yace.compiler import Compiler
from yace.parser import c_to_yace
from yace.idl.formater import do_format


def parse_args():
    """Parse command-line arguments"""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "filepath",
        nargs="+",
        type=Path,
        help="Path to Yace File or C Header",
    )
    parser.add_argument(
        "--target",
        nargs="+",
        choices=[target.NAME for target in Compiler.TARGETS],
        default=[target.NAME for target in Compiler.TARGETS][0],
        help="Targets",
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=Path("output"),
        help="Path to output directory, for emitted code / artifacts",
    )
    parser.add_argument(
        "--log-level",
        "-l",
        action="append_const",
        const=1,
        default=[],
        help="Increase log-level.",
    )

    parser.add_argument(
        "--c-to-yace",
        action="store_true",
        help="Treat filepath as C Header and use it to emit a Yace-file, then exit",
    )
    parser.add_argument(
        "--lint",
        action="store_true",
        help="Parse and check the given Yace-files, then exit",
    )
    parser.add_argument(
        "--format",
        action="store_true",
        help="Format the given Yace-file, then exit",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {version}",
    )

    return parser.parse_args()


def main():
    """Emit enums, structs, and pretty-printer functions for them"""

    ok = False
    try:
        args = parse_args()
        args.filepath = [filepath.resolve() for filepath in args.filepath]

        levels = [log.ERROR, log.INFO, log.DEBUG]
        log.basicConfig(
            format="%(levelname)s:%(module)s:%(funcName)s(): %(message)s",
            level=levels[min(sum(args.log_level), len(levels) - 1)],
        )

        if args.format:  # Yace-file formater
            return sys.exit(do_format(args.filepath))

        if args.c_to_yace:  # C to Yace-file Compiler
            return sys.exit(c_to_yace(args.filepath, args.output))

        yace = Compiler(
            args.target, args.output
        )  # Yace-file linter and/or Yace-file Compiler
        ok = all(
            [
                yace.process(path, ["parse", "lint"] if args.lint else Compiler.STAGES)
                for path in args.filepath
            ]
        )
    except Exception as exc:
        log.error("Unhandled Exception: message(%s)", exc)
        log.error("Unhandled Exception: increase log-level (-ll) for trace")
        log.error("Unhandled Exception(%s)", exc, exc_info=True)

    return sys.exit(0 if ok else 1)



#!/usr/bin/env python3
"""
The **yace** function types consists of:

* Function Declarations: :class:`.Function`
* Function Pointer Declarations: :class:`.FunctionPointer`
* Both of the above make use of

  * :class:`.Parameter`
  * :class:`.ReturnType`

Their idl representation follows below.
"""
import typing

from .base import Documented, Entity, Named, Typed


class ReturnType(Entity, Documented, Typed):
    """
    Function return type
    """

    cls: str = "ret"


class Parameter(Entity, Named, Documented, Typed):
    """
    Function parameter
    """

    cls: str = "param"


class Function(Entity, Named, Documented):
    """
    Function declarations
    """

    cls: str = "fun"
    ret: ReturnType

    parameters: typing.List[Parameter] = []


class FunctionPointer(Entity, Named, Documented, Typed):
    """
    Function pointer declarations
    """

    cls: str = "fun_ptr"
    ret: ReturnType

    parameters: typing.List[Parameter]



"""
What to do about "static" and "const"? The filter_typedecl.template currently
ignores them, except for the special-case "c_char_p".

"""
import copy
import logging as log
import shutil
from pathlib import Path

from yace.emitters import Emitter
from yace.errors import TransformationError
from yace.targets.target import Target
from yace.tools import Black, Isort, Python3
from yace.transformations import Camelizer


class Ctypes(Target):
    """
    Several helper functions
    """

    NAME = "ctypes"

    def __init__(self, output):
        super().__init__(Ctypes.NAME, output)

        self.emitter = Emitter(self.output, Ctypes.NAME)

        self.tools = {
            "black": Black(self.output),
            "isort": Isort(self.output),
            "python": Python3(self.output),
        }

    def transform(self, model):
        """
        Transform the given idl

        * Transform symbols according to :class:`yace.transformation.CStyle`

        That it currently the only thing done to the **yace** IDL.
        """

        transformed = copy.deepcopy(model)

        status = Camelizer(transformed).walk()
        if not all([res for res in status]):
            raise TransformationError("The CStyle transformation failed")

        return transformed

    def emit(self, model):
        """Emit code"""

        # Copy the generic ctypes-sugar from resources
        sugar_path = (self.output / "ctypes_sugar.py").resolve()
        shutil.copyfile(Path(__file__).parent / sugar_path.name, sugar_path)
        self.sources.append(sugar_path)

        # Generate the bindings / Python API
        files = [
            ((self.output / f"{model.meta.prefix}.py").resolve(), "file_api"),
            ((self.output / f"{model.meta.prefix}_check.py").resolve(), "file_check"),
        ]
        for path, template in files:
            with path.open("w") as file:
                file.write(
                    self.emitter.render(
                        template,
                        {
                            "meta": model.meta,
                            "entities": model.entities,
                            "headers": self.headers,
                        },
                    )
                )
            self.sources.append(path)

    def format(self):
        """
        Run 'black' and 'isort' on self.sources
        """

        for tool in ["black", "isort"]:
            self.tools[tool].run([str(path) for path in self.sources])

    def check(self):
        """Build generated sources and run the generated test-program"""

        log.info("Not there yet...")



import copy
import logging as log
import typing
from pathlib import Path

from yace.idl.linter import Linter
from yace.model import Model
from yace.targets.capi.target import CAPI
from yace.targets.ctypes.target import Ctypes


class Compiler(object):
    """
    Encapsulation of **yace** compiler stages

    * Parse
    * Lint
    * Transform
    * Emit
    * Format
    * Check

    The first two stages are generic, the third is generic however is usually
    performed for target-specific-reasons such as re-structuring the IDL to
    make the code-emitter simpler, the last three are target-specific.
    """

    STAGES = ["parse", "lint", "transform", "emit", "format", "check"]
    TARGETS = [CAPI, Ctypes]

    def __init__(self, targets: typing.List[str], output: Path):
        self.targets = [target for target in Compiler.TARGETS if target.NAME in targets]
        self.output = output.resolve()

    def process(
        self, path: Path, stages: typing.Optional[typing.List[str]] = None
    ) -> bool:
        """
        Take 'path' through the given compiler 'stages'
        """

        if stages is None:
            stages = Compiler.STAGES

        log.info("Path: '%s', stages: '%s'", path, stages)
        self.output.mkdir(parents=True, exist_ok=True)

        log.info("Stage: 'parse'")
        model_orig = Model.from_path(path)

        if "lint" in stages:
            log.info("Stage: 'lint'")
            nerrors = Linter(model_orig).check()
            if nerrors:
                log.error("Linter-errors: see above/log; stopping.")
                return False

        targets = [cls(self.output) for cls in self.targets]
        if not all([tgt.is_ready() for tgt in targets]):
            log.error("One or more targets !ready; see above/log. stopping.")
            return False

        for target in targets:
            log.info("Target: %s", target.NAME)

            model = copy.deepcopy(model_orig)

            if "transform" in stages:
                log.info("Stage: 'transform'")
                model = target.transform(model)

            if "emit" in stages:
                log.info("Stage: 'emit'")
                err = target.emit(model)
                if err:
                    log.error("Got error, stopping.")
                    break

            if "format" in stages:
                log.info("Stage: 'format'")
                err = target.format()
                if err:
                    log.error("Got error, stopping.")
                    break

            if "check" in stages:
                log.info("Stage: 'check'")
                err = target.check()
                if err:
                    log.error("Got error, stopping.")
                    break

        return True



"""
The :class:`.CAPI` emits a C API consisting of the following files:

* ``lib{meta.prefix}.h``        -- Library header bundle
* ``lib{meta.prefix}_core.h``   -- The core API declarations
* ``lib{meta.prefix}_pp.h``     -- Pretty-printer definitions
* ``{meta.prefix}_pp.c``        -- Pretty-printer implementation

additionally then:

* ``{meta.prefix}_check.c``     -- ``main()`` program checking generated code

This program is utilized by the 'check-stage' and is as such not part of the
emitted C API but rather a test of it.

Last then:

* ``doxygen.conf``              -- Doxygen configuration for the above

Is emitted, this is to produce DoxyGen documentation for API.

The 'format' stage also emits the files:

* ``hdr.clang-format`` -- clang-format rule-file for header-files
* ``src.clang-format`` -- clang-format rule-file for source-files

Thus, the above files are what you should expect to see in the output-directory
"""
import copy
import logging as log
import shutil
from pathlib import Path

from yace.emitters import Emitter
from yace.errors import TransformationError
from yace.targets.target import Target
from yace.tools import ClangFormat, Doxygen, Gcc
from yace.transformations import CStyle


class CAPI(Target):
    """
    Several helper functions
    """

    NAME = "capi"
    CFLAGS = ["-std=c99", "-pedantic-errors", "-Wall", "-Werror"]

    def __init__(self, output):
        super().__init__(CAPI.NAME, output)

        self.emitter = Emitter(self.output, CAPI.NAME)

        self.tools["clang-format"] = ClangFormat(self.output)
        self.tools["doxygen"] = Doxygen(self.output)
        self.tools["gcc"] = Gcc(self.output)

    def transform(self, model):
        """
        Transform the given idl

        * Transform symbols according to :class:`yace.transformation.CStyle`

        That it currently the only thing done to the **yace** IDL.
        """

        transformed = copy.deepcopy(model)

        status = CStyle(transformed).walk()
        if not all([res for res in status]):
            raise TransformationError("The CStyle transformation failed")

        return transformed

    def emit(self, model):
        """Emit code"""

        files = [
            (f"lib{model.meta.prefix}_core.h", "file_core.h", self.headers),
            (f"lib{model.meta.prefix}_pp.h", "file_pp.h", self.headers),
            (f"lib{model.meta.prefix}.h", "file_bundle.h", self.headers),
            (f"{model.meta.prefix}_pp.c", "file_pp.c", self.sources),
            (f"{model.meta.prefix}_check.c", "file_check.c", self.sources),
            ("doxygen.conf", "doxygen", self.aux),
        ]
        for filename, template, container in files:
            path = (self.output / filename).resolve()
            with (path).open("w") as file:
                content = self.emitter.render(
                    template,
                    {
                        "meta": model.meta,
                        "entities": model.entities,
                        "headers": self.headers,
                    },
                )
                if content[-1] != "\n":
                    content += "\n"
                file.write(content)

            log.info("produced: %s", path)
            container.append(path)

    def format(self):
        """
        Transfer the clang-format definition from package to output and
        invokes the clang-formater
        """

        path = Path(__file__).parent
        for rules, container in [
            (ClangFormat.CLANGFORMAT_STYLE_C, self.sources),
            (ClangFormat.CLANGFORMAT_STYLE_H, self.headers),
        ]:
            shutil.copyfile(path / rules, self.output / rules)
            self.tools["clang-format"].run(
                [f"--style=file:{rules}", "-i"] + [str(f) for f in container],
            )

        self.tools["doxygen"].run([Doxygen.DOXYGEN_CONF])

    def check(self):
        """Build generated sources and run the generated test-program"""

        self.tools["gcc"].run(
            CAPI.CFLAGS + ["-I", str(self.output)] + [str(p) for p in self.sources]
        )



"""
These are Yace Interface Definition transformations, transforming the loaded
definitions into a different representation, this is usually done in order to
produce an equivalent model with a representation better suited for emitting
code.

E.g. symbols are snake_case in Yace Files, however, when emitting Rust or Python
code, then it is very useful to have this transformed to CamelCase. By doings
so at the IDL-level, then the code-emitters need not concern inself with
UPPERCASE/lowercase/snake_case/camelCase/PascalCase etc. this makes emitters a
lot simpler.

Also, at the IDL-level, then coding-conventions such as CamelCase can be
shared by multiple targets.
"""
from yace.emitters import camelcase
from yace.idl.derivedtypes import Struct, Union
from yace.model import ModelWalker


class CStyle(ModelWalker):
    """
    Transform symbols:

    * All ``.sym`` in top-level entities are prefixed with ``meta.prefix_``
    * :class:`yace.idl.constants.Define` -> upper()
    * :class:`yace.idl.constants.EnumValue` -> upper()

    This transformation assumes a valid **Yace**-file according to the linter.
    Thus, the symbols are all lower-snake_case, and transformation not
    described above is not performed.
    """

    def visit(self, current, ancestors, depth):
        """..."""

        if "sym" not in current.all:
            return (True, None)

        if depth == 0:
            current.sym = f"{self.model.meta.prefix}_{current.sym}"
            pass

        if current.cls in ["define"]:
            current.sym = current.sym.upper()
        elif current.cls in ["enum_value"]:
            current.sym = ancestors[-1].sym.upper() + "_" + current.sym.upper()

        return (True, None)


class Camelizer(ModelWalker):
    """
    Transform symbols:

    * Enum -> camelize()
    * EnumValue -> upper()
    * Struct|Enum|Bitfield -> camelize()

    Additionally, then Enum / Struct / Union is added  "postfix" to
    ``entity.sym``. This is done for non-C languages, as enum/struct/union
    C-types rarely has equivalents in "foreign" languages. Rather, these are
    objects with subclasses encapsulating them, e.g. the Python ctypes.

    Thus, the Enum/Struct/Union needs to be added to the identifier, since it
    will otherwise potentially collide with other entities.

    This transformation assumes a valid **Yace**-file according to the linter.
    Thus, the symbols are all lower-snake_case, and transformation not
    described above is not performed.
    """

    def visit(self, current, ancestors, depth):
        if "sym" not in current.all:
            return (True, None)

        if current.cls in ["define"]:
            current.sym = current.sym.upper()
        elif current.cls in ["enum", "struct", "union"]:
            current.sym = "_".join([current.sym, current.cls])
            current.sym = camelcase(current.sym)
        elif current.cls in ["enum_value"]:
            current.sym = current.sym.upper()

        return (True, None)


class HoistAnonMembers(ModelWalker):
    """
    Extract nested entities, that is, struct/union/bitfield which from nested
    to top-level declaration, leaving behind a replace of the entity with a
    field with a typespec matching the extracted entity"""

    def visit(self, current, ancestors, depth):
        if depth != 0:
            return (True, None)

        if type(current) not in [Struct, Union]:
            return (True, None)

        # TODO: top-level struct or union



"""
Addition to ctypes, making it slightly nicer to use, including:

* gen_search_paths(libname)
  * find_library() + pkg-config

* load(libname)

* typedef additions
  * c_log_double_t
  * c_uint128
  * c_int128
  * void

TODO

* Dict-accessor for structs / unions
* Convenient typecast-helpers

Copyright (c) 2023 Simon A. F. Lund <os@safl.dk>
SPDX-License-Identifier: BSD-3-Clause
"""
import ctypes
import ctypes.util
import os
import platform
import subprocess

SHARED_EXT = {
    "linux": "so",
    "windows": "dll",
    "darwin": "dylib",
}

c_int128 = ctypes.c_ubyte * 16
c_uint128 = c_int128
void = None
if ctypes.sizeof(ctypes.c_longdouble) == 8:
    c_long_double_t = ctypes.c_longdouble
else:
    c_long_double_t = ctypes.c_ubyte * 8


class Enum(object):
    """Encapsulation of C enum"""

    pass


class Structure(ctypes.Structure):
    """Encapsulation of C structs"""

    pass


class Union(ctypes.Union):
    """Encapsulation of C union"""

    pass


def gen_search_paths(libname):
    """
    Yields search-paths for the shared library with the given 'libname'. It is
    an extension of ``ctypes.util.find_library()`` using ``pkg-config``.
    """

    path = ctypes.util.find_library(libname)
    if path:
        yield path

    try:
        proc = subprocess.run(
            ["pkg-config", libname, "--variable=libdir"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if not proc.returncode:
            ext = SHARED_EXT.get(platform.system().lower(), "so")

            yield os.path.join(
                proc.stdout.decode("utf-8").strip(), f"lib{libname}.{ext}"
            )
    except subprocess.CalledProcessError:
        pass


def load(libname):
    """Dynamically load the shared library named 'libname'"""

    for spath in gen_search_paths(libname):
        try:
            lib = ctypes.CDLL(spath)
            if lib:
                return lib
        except OSError:
            continue

    return None



#!/usr/bin/env python3
"""
The **yace** constants consists of:

* Constant Values:

  * :class:`yace.idl.constants.String`
  * :class:`yace.idl.constants.Dec`
  * :class:`yace.idl.constants.Hex`

* Symbolic Constants:

  * :class:`yace.idl.constants.Define`

* Enum Types:

  * :class:`yace.idl.constants.Enum`
  * :class:`yace.idl.constants.EnumValue`

Their idl representation follows below.
"""
import typing

from .base import Documented, Entity, Named


class String(Entity):
    """
    Representation of a literal string e.g. in C::

        "this is a string-literal"

    """

    cls: str = "str"
    lit: str

    def is_valid_lit(self):
        if not isinstance(self.lit, str):
            return False, f"Unsupported lit[{self.lit}]"

        return True, None


class Hex(Entity):
    """
    Representation of a hexadecimal constant

    The emitted C code should be hexadecimal notation of literal integer
    values, such as those used by :class:`yace.idl.constants.Define` and
    :class:`yace.idl.enumtypes.Enum`::

        #define FOO 0xACDC

    """

    cls: str = "hex"
    lit: int

    def is_valid_lit(self):
        if not isinstance(self.lit, int):
            return False, f"Unsupported lit[{self.lit}]"

        return True, None


class Dec(Entity):
    """Representation of a an integer literal in decimal"""

    cls: str = "dec"
    lit: int

    def is_valid_lit(self):
        if not isinstance(self.lit, int):
            return False, f"Unsupported lit[{self.lit}]"

        return True, None


class Define(Entity, Named, Documented):
    """
    C MACROS, one of biggest headaches when it comes to FFI interfaces,
    especially for interpreted languages such as Python, since
    MACRO-expressions have no symbols to "call" as these are all
    pre-preprocossed and gone thus, no longer available.

    Thus, **yace** only support the definition of MACRO values, such as::

        #define SOME_GLOBALLY_CONSTANT 0xACDC

    Definitions such as these to translate fairly when into equivalent globally
    available and constant constructs in other languages
    """

    cls: str = "define"
    val: typing.Any


class EnumValue(Entity, Named, Documented):
    """
    Representation of values in :class:`yace.idl.constants.Enum`.

    By default, the C emitter produces::

        SYMBOL = lit.val

    for the :class:`yace.idl.constants.EnumValue`.
    """

    cls: str = "enum_value"
    val: typing.Any = None

    def is_valid_val(self):
        """Checks whether the :class:`yace.idl.constants.EnumValue` is valid"""

        if not isinstance(self.val, (Dec, Hex)):
            return (
                False,
                f"EnumValue[{self.sym}][Unsupported Instance]: {self.val}",
            )

        return (True, None)


class Enum(Entity, Named, Documented):
    """
    Representation of enumeration values, note that literals on the
    :class:`yace.idl.constants.EnumValue` are optional, an
    :class:`yace.idl.constants.Enum` with **non** literal
    :class:`yace.idl.constants.EnumValue` looks like::

        enum example {
            FOO,
            BAR,
            BAZ,
        };

    Whereas, an :class:`yace.idl.constants.Enum` with **literal**
    :class:`yace.idl.constants.EnumValue` looks like::

        enum example {
            FOO = 0xACDC,
            BAR = 0x1337,
            BAZ = 0xBEEF,
        };

    for the :class:`yace.idl.constants.EnumValue``.
    """

    cls: str = "enum"
    members: typing.List[EnumValue]

    def is_valid_members(self):
        """Checks whether the :class:`yace.idl.constants.Enum` is valid"""

        invalid_types = [type(m) for m in self.members if not isinstance(m, EnumValue)]
        if len(invalid_types):
            return (
                False,
                f"Invalid types in {self.sym}.members; "
                "expecting: EnumValue, got: [{invalid_types}]",
            )

        if String in [type(m) for m in self.members]:
            return (False, "Enum-values must be numeral-types")

        return (True, None)



"""

"""
import os
from abc import ABC, abstractmethod


class Target(ABC):
    """
    Encapsulation of a **yace** Emitter Target (yet). This serves as
    documentation for what is needed when adding a target to **yace**.
    Additionally, the Target-constructur sets up commonly used attributes of
    the Target and creates the sub-directory for artifacts produced by the
    target. Labelled, as artifacts, as it is not just for the emitted code, but
    also log-files from running tools, and other similar side-effects.
    """

    def __init__(self, name, output):
        self.name = name  # Name of the compiler-target
        self.output = output / name  # Location to emit code to

        self.headers = []  # Resolved paths to emitted headers
        self.sources = []  # Resolved paths to emitted sources
        self.aux = []  # Resolved paths to auxilary files e.g. Doxy Conf

        self.tools = {}  # Dictionary of :class:`yace.tools.Tool` instances

        os.makedirs(self.output, exist_ok=True)

    def is_ready(self):
        """Pre-flight check, inspect availability of self.tools"""

        return all([tool.exists() for label, tool in self.tools.items()])

    @abstractmethod
    def transform(self, model):
        """
        Transform the given 'idl' for code-emission.
        """

    @abstractmethod
    def emit(self, model):
        """
        Emit code for the given idl, using your weapons of choice, common
        choice would be to utilize an instance of the class:`.Emitter`.
        """

    @abstractmethod
    def format(self):
        """
        Format the emitted source-code, e.g. call tools such as clang-format,
        black, isort, rustfmt, etc. depending of what applies to the generated
        code.
        """

    @abstractmethod
    def check(self):
        """
        Compile the generated code and execute it. Or in case of an interpreted
        language just execute something verifying the generated code. This
        could for example be a test-program emitted earlier.
        """
